////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Geometry programs using Cg
// sgreen 5/2006
//
// Email: sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
////////////////////////////////////////////////////////////////////////////////////////////////////


#include <stdio.h>
#include <stdlib.h>
#include <map>

#include <GL/glew.h>
#include <GL/glut.h>

#include <windows.h>
#include <GL/gl.h>

#include <nvGlutManipulators.h>
#include <nvModel.h>
#include <nvGlutWidgets.h>
#include <nvSDKPath.h>

#include <Cg/cg.h>
#include <Cg/cgGL.h>

////////////////////////////////////////////////////////////////////////////////
//
//  globals
//
////////////////////////////////////////////////////////////////////////////////

#define DEMO_PATH "src/cg_geometry_program/shaders/"
using std::map;\

enum UIOption {
    OPTION_DISPLAY_WIREFRAME,
    OPTION_OVERLAY_OBJECT,
    OPTION_ANIMATE,
    OPTION_USE_PROGRAM,
    OPTION_COUNT,
	OPTION_DRAW_MODEL
};
bool options[OPTION_COUNT];
map<char,UIOption> optionKeyMap;
nv::GlutExamine manipulator;

nv::GlutUIContext ui;

enum ModelMode {
	MODEL_NONE,
	MODEL,
	MODEL_ADJ,
	MODEL_POINTS,
	MODEL_EDGES
};

bool enableBlending = false;
bool enableDepthTest = true;

//
// model data
//
nv::Model *model = 0;
char *model_filename = "media/models/cow.obj";
nv::vec3f modelMin, modelMax;

// Path finding utility
nv::SDKPath sdkPath;

// window size
int width = 512, height = 512;

// level of subdivision on tesselation shaders
int segments = 10;
int passes = 10;

GLuint rand_tex, noise_tex;

CGcontext cg_context; 
CGprofile cg_vprofile, cg_gprofile, cg_fprofile;

struct Program {
    const char *name;
    const char *filename;
    const char *vprog_entry, *gprog_entry, *fprog_entry;    // entry points for vertex, geometry and fragment programs
    GLenum prim;
	ModelMode modelmode;
	bool multipass;
    CGprogram vprog, gprog, fprog;
};

Program prog[] = {
    { "passthru", DEMO_PATH "passthru.cg", "transform_light_vp", "passthru_gp", 0, GL_TRIANGLES, MODEL, false, 0, 0, 0 },
    { "shrink", DEMO_PATH "passthru.cg", "transform_light_vp", "shrink_gp", 0, GL_TRIANGLES, MODEL, false, 0, 0, 0 },
    { "bezier", DEMO_PATH "curves.cg", "transform_vp", "bezier_gp", 0, GL_LINES_ADJACENCY_EXT, MODEL_NONE, false, 0, 0, 0 },
    { "bspline", DEMO_PATH "curves.cg", "transform_vp", "bspline_gp", 0, GL_LINE_STRIP_ADJACENCY_EXT, MODEL_NONE, false, 0, 0, 0 },
    { "extrude", DEMO_PATH "curves.cg", "passthru_vp", "extrude_gp", "extrude_fp", GL_LINE_STRIP_ADJACENCY_EXT, MODEL_NONE, false, 0, 0, 0 },
    { "hair", DEMO_PATH "curves.cg", "passthru_vp", "hair_gp", "hair_fp", GL_TRIANGLES, MODEL, true, 0, 0, 0 },
    { "fins", DEMO_PATH "fins.cg", "passthru_vp", "fins_gp", 0, GL_TRIANGLES, MODEL, false, 0, 0, 0 },
    { "silhouette", DEMO_PATH "silhouette.cg", "silhouette_vp", "silhouette_gp", 0, GL_TRIANGLES_ADJACENCY_EXT, MODEL_ADJ, false, 0, 0, 0 },
    { "rasterize bbox", DEMO_PATH "rasterize.cg", "rasterize_vp", "rasterize_gp", 0, GL_TRIANGLES, MODEL, false, 0, 0, 0 },
    { "rasterize", DEMO_PATH "rasterize.cg", "rasterize_vp", "rasterize_gp", "rasterize_fp", GL_TRIANGLES, MODEL, false, 0, 0, 0 },
};

#define NPROGS sizeof(prog) / sizeof(Program)

#define CHECK_ERRORS  \
{ \
    GLenum err = glGetError(); \
    if (err) \
        printf( "Error %x at line %d\n", err, __LINE__); \
}

Program *current_prog = &prog[7];
int current_prog_idx = 7;

////////////////////////////////////////////////////////////////////////////////
//
//  Functions
//
////////////////////////////////////////////////////////////////////////////////

GLuint create_rand_texture(GLenum target, GLenum format, int w, int h);
GLuint create_noise_texture(GLenum format, int w, int h, int d);

//////////////////////////////////////////////////////////////////////
//
// Callback function used for reporting Cg compile errors
//
//////////////////////////////////////////////////////////////////////
void cgErrorCallback()
{
    CGerror lastError = cgGetError();
    if(lastError)
    {
        printf("%s\n", cgGetErrorString(lastError));
        printf("%s\n", cgGetLastListing(cg_context));
        exit(1);
    }
}

//////////////////////////////////////////////////////////////////////
//
//  Function for setting up initial OpenGL state. Also compiles Cg
// programs.
//
//////////////////////////////////////////////////////////////////////
void init_opengl()
{
    glewInit();
    
    if (!glewIsSupported(
        "GL_VERSION_2_0 "
        "GL_ARB_vertex_program "
        "GL_ARB_fragment_program "
        "GL_NV_gpu_program4 " // includes GL_NV_geometry_program4
        "GL_NV_texture_rectangle "
        "GL_ARB_texture_float "
		))
    {
        printf("Unable to load necessary extension(s)\n");
        printf("This sample requires:\n  OpenGL 2.0\n  GL_ARB_vertex_program\n  GL_ARB_fragment_program\n"
               "GL_NV_gpu_program4\n  GL_NV_texture_rectangle\n  GL_ARB_texture_float\nExiting...\n");
        exit(-1);
    }

    // create some helper textures
    rand_tex = create_rand_texture(GL_TEXTURE_RECTANGLE_NV, GL_RGBA16F_ARB, 256, 256);
    noise_tex = create_noise_texture(GL_RGBA16F_ARB, 16, 16, 16);

    // get geometry program limits
    GLint max_output_vertices, max_total_output_components;
    glGetProgramivARB(GL_GEOMETRY_PROGRAM_NV, GL_MAX_PROGRAM_OUTPUT_VERTICES_NV, &max_output_vertices);
    glGetProgramivARB(GL_GEOMETRY_PROGRAM_NV, GL_MAX_PROGRAM_TOTAL_OUTPUT_COMPONENTS_NV, &max_total_output_components);
    printf("max output vertices = %d\n", max_output_vertices);
    printf("max total output components = %d\n", max_total_output_components);

    // set normal rendering options
    glEnable(GL_DEPTH_TEST);
    glDepthFunc( GL_LEQUAL);
//    glClearColor(0.5f, 0.5f, 0.5f, 1.0f);
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

    glDisable(GL_CULL_FACE);

    // Create Cg Context
    cg_context = cgCreateContext();
    cgSetErrorCallback(cgErrorCallback);

    // load Cg programs
    cg_vprofile = cgGLGetLatestProfile(CG_GL_VERTEX);
    cg_gprofile = cgGLGetLatestProfile(CG_GL_GEOMETRY);
    cg_fprofile = cgGLGetLatestProfile(CG_GL_FRAGMENT);

    const char *args[] = {
        "-unroll",
        "all",
        0,
    };

    // load programs
    for(int i=0; i<NPROGS; i++) {
        Program &p = prog[i];
        std::string resolved_path;

        if ( sdkPath.getFilePath( p.filename, resolved_path)) {
            p.vprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_vprofile, p.vprog_entry, 0);
            cgGLLoadProgram(p.vprog);

            p.gprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_gprofile, p.gprog_entry, args);
            cgGLLoadProgram(p.gprog);

            if (p.fprog_entry) {
                p.fprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_fprofile, p.fprog_entry, 0);
                cgGLLoadProgram(p.fprog);
            }
        }
        else {
            fprintf( stderr, "Failed to find shader file '%s'\n", p.filename);
        }

        
    }

    glutReportErrors();
}

//////////////////////////////////////////////////////////////////////
//
//  Create a random float between 0 - 1 inclusive
//
//////////////////////////////////////////////////////////////////////
float frand()
{
    return rand() / (float) RAND_MAX;
}

//////////////////////////////////////////////////////////////////////
GLuint create_rand_texture(GLenum target, GLenum format, int w, int h)
{
  float *data = new float [w*h*4];
  float *ptr = data;
  for(int i=0; i<w*h; i++) {
    *ptr++ = frand();
    *ptr++ = frand();
    *ptr++ = frand();
    *ptr++ = frand();
  }

  GLuint texid;
  glGenTextures(1, &texid);
  glBindTexture(target, texid);

  glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_REPEAT);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexImage2D(target, 0, format, w, h, 0, GL_RGBA, GL_FLOAT, data);

  return texid;
}

//////////////////////////////////////////////////////////////////////
GLuint create_noise_texture(GLenum format, int w, int h, int d)
{
  float *data = new float [w*h*d*4];
  float *ptr = data;
  for(int i=0; i<w*h*d; i++) {
    *ptr++ = frand()*2.0-1.0;
    *ptr++ = frand()*2.0-1.0;
    *ptr++ = frand()*2.0-1.0;
    *ptr++ = frand()*2.0-1.0;
  }

  GLuint texid;
  glGenTextures(1, &texid);
  GLenum target = GL_TEXTURE_3D;
  glBindTexture(target, texid);

  glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(target, GL_TEXTURE_WRAP_R, GL_REPEAT);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexImage3D(target, 0, format, w, h, d, 0, GL_RGBA, GL_FLOAT, data);

  return texid;
}

//////////////////////////////////////////////////////////////////////
void DrawTriangles( const nv::Model *model) {
    nv::vec3f center = (modelMax + modelMin) * 0.5f;
    float factor = 2.0f / nv::length(modelMax - modelMin);

    CHECK_ERRORS;

    glMatrixMode( GL_MODELVIEW);
    glPushMatrix();
    glScalef( factor, factor, factor);
    glTranslatef( -center.x, -center.y, -center.z);
    glVertexPointer( model->getPositionSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices());
    glNormalPointer( GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledNormalOffset());

    CHECK_ERRORS;

    glEnableClientState( GL_VERTEX_ARRAY);
    glEnableClientState( GL_NORMAL_ARRAY);

    if ( model->hasTexCoords()) {
        glTexCoordPointer( model->getTexCoordSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledTexCoordOffset());
        glEnableClientState( GL_TEXTURE_COORD_ARRAY);
    }    
    if ( model->hasTangents()) {
		glClientActiveTexture(GL_TEXTURE1);
        glTexCoordPointer( model->getTangentSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledTangentOffset());
        glEnableClientState( GL_TEXTURE_COORD_ARRAY);
    }   

    CHECK_ERRORS;

    glDrawElements( GL_TRIANGLES, model->getCompiledIndexCount( nv::Model::eptTriangles), GL_UNSIGNED_INT, model->getCompiledIndices( nv::Model::eptTriangles));

    glDisableClientState( GL_VERTEX_ARRAY);
    glDisableClientState( GL_NORMAL_ARRAY);
	glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState( GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState( GL_TEXTURE_COORD_ARRAY);

    glPopMatrix();
}

//////////////////////////////////////////////////////////////////////
void DrawPoints( const nv::Model *model) {
    nv::vec3f center = (modelMax + modelMin) * 0.5f;
    float factor = 2.0f / nv::length(modelMax - modelMin);

    CHECK_ERRORS;

    glMatrixMode( GL_MODELVIEW);
    glPushMatrix();
    glScalef( factor, factor, factor);
    glTranslatef( -center.x, -center.y, -center.z);
    glVertexPointer( model->getPositionSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices());
    glNormalPointer( GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledNormalOffset());

    CHECK_ERRORS;

    glEnableClientState( GL_VERTEX_ARRAY);
    glEnableClientState( GL_NORMAL_ARRAY);

    if ( model->hasTexCoords()) {
        glTexCoordPointer( model->getTexCoordSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledTexCoordOffset());
        glEnableClientState( GL_TEXTURE_COORD_ARRAY);
    }    

    CHECK_ERRORS;

    glDrawElements( GL_POINTS, model->getCompiledIndexCount( nv::Model::eptPoints), GL_UNSIGNED_INT, model->getCompiledIndices( nv::Model::eptPoints));

    glDisableClientState( GL_VERTEX_ARRAY);
    glDisableClientState( GL_NORMAL_ARRAY);
    glDisableClientState( GL_TEXTURE_COORD_ARRAY);

    glPopMatrix();
}

//////////////////////////////////////////////////////////////////////
void DrawEdges( const nv::Model *model) {
    nv::vec3f center = (modelMax + modelMin) * 0.5f;
    float factor = 2.0f / nv::length(modelMax - modelMin);

    CHECK_ERRORS;

    glMatrixMode( GL_MODELVIEW);
    glPushMatrix();
    glScalef( factor, factor, factor);
    glTranslatef( -center.x, -center.y, -center.z);
    glVertexPointer( model->getPositionSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices());
    glNormalPointer( GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledNormalOffset());

    CHECK_ERRORS;

    glEnableClientState( GL_VERTEX_ARRAY);
    glEnableClientState( GL_NORMAL_ARRAY);

    if ( model->hasTexCoords()) {
        glTexCoordPointer( model->getTexCoordSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledTexCoordOffset());
        glEnableClientState( GL_TEXTURE_COORD_ARRAY);
    }    

    CHECK_ERRORS;

    glDrawElements( GL_LINES, model->getCompiledIndexCount( nv::Model::eptEdges), GL_UNSIGNED_INT, model->getCompiledIndices( nv::Model::eptEdges));

    glDisableClientState( GL_VERTEX_ARRAY);
    glDisableClientState( GL_NORMAL_ARRAY);
    glDisableClientState( GL_TEXTURE_COORD_ARRAY);

    glPopMatrix();
}

//////////////////////////////////////////////////////////////////////
void DrawWithAdjacency( const nv::Model *model) {
    nv::vec3f center = (modelMax + modelMin) * 0.5f;
    float factor = 2.0f / nv::length(modelMax - modelMin);

    glMatrixMode( GL_MODELVIEW);
    glPushMatrix();
    glScalef( factor, factor, factor);
    glTranslatef( -center.x, -center.y, -center.z);
    glVertexPointer( model->getPositionSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices());
    glNormalPointer( GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledNormalOffset());

    glEnableClientState( GL_VERTEX_ARRAY);
    glEnableClientState( GL_NORMAL_ARRAY);

    CHECK_ERRORS;

    if ( model->hasTexCoords()) {
		glClientActiveTexture(GL_TEXTURE0);
        glTexCoordPointer( model->getTexCoordSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledTexCoordOffset());
        glEnableClientState( GL_TEXTURE_COORD_ARRAY);
    }    
    if ( model->hasTangents()) {
		glClientActiveTexture(GL_TEXTURE1);
        glTexCoordPointer( model->getTangentSize(), GL_FLOAT, model->getCompiledVertexSize() * sizeof(float), model->getCompiledVertices() + model->getCompiledTangentOffset());
        glEnableClientState( GL_TEXTURE_COORD_ARRAY);
    }

    glDrawElements( GL_TRIANGLES_ADJACENCY_EXT, model->getCompiledIndexCount( nv::Model::eptTrianglesWithAdjacency), GL_UNSIGNED_INT, model->getCompiledIndices( nv::Model::eptTrianglesWithAdjacency));

    CHECK_ERRORS;
    glDisableClientState( GL_VERTEX_ARRAY);
    glDisableClientState( GL_NORMAL_ARRAY);
	glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState( GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState( GL_TEXTURE_COORD_ARRAY);

    glPopMatrix();

}

//////////////////////////////////////////////////////////////////////
void draw_prim(GLenum prim)
{
    float v[][2] = {
        { 0.0, 0.0 },
        { 0.5, -0.2 },
        { 1.0, 0.4 },
        { 0.2, 0.8 },
        { -0.2, 1.0 },
        { -0.5, 0.0 },
        { -0.9, 0.3 }
    };
    int nverts = 7;

    switch(prim) {
    case GL_POINTS:
        glBegin(GL_POINTS);
        glVertex2f(0.5, 0.5);
        glVertex2f(-0.2, 0.1);
        glVertex2f(0.7, -0.2);
        glEnd();
        CHECK_ERRORS;
        break;

    case GL_LINES:
        glBegin(GL_LINES);
        for(int i=0; i<nverts-1; i++) {
            glVertex2fv(v[i]);
            glVertex2fv(v[i+1]);
        }
        glEnd();
        CHECK_ERRORS;
        break;

    case GL_TRIANGLES:
        glNormal3f(0.0, 0.0, 1.0);
        glMultiTexCoord3f(GL_TEXTURE1, 0.0, 1.0, 0.0);

        glBegin(GL_TRIANGLES);
        glMultiTexCoord1f(GL_TEXTURE3, 0.0);
        glColor3f(1.0, 0.0, 0.0); glVertex3f(0.0, 0.0, 0.0);
        glColor3f(0.0, 1.0, 0.0); glVertex3f(1.0, 0.0, 0.0);
        glColor3f(0.0, 0.0, 1.0); glVertex3f(1.0, 1.0, 0.0);

        glMultiTexCoord1f(GL_TEXTURE3, 1.0);
        glColor3f(1.0, 0.0, 0.0); glVertex3f(1.0, 1.1, 1.0);
        glColor3f(0.0, 1.0, 0.0); glVertex3f(1.0, 0.0, 0.5);
        glColor3f(0.0, 0.0, 1.0); glVertex3f(0.0, 0.0, -0.5);
        glEnd();

        CHECK_ERRORS;
        break;

    case GL_TRIANGLES_ADJACENCY_EXT:
        {
            // 1 2 3
            //   0 4
            //     5
            float v[][3] = {
                0.0, 0.0, 0.0,
                -1.0, 1.0, -1.0,
                0.0, 1.0, 0.0,
                1.0, 1.0, -1.0,
                1.0, 0.0, 0.0,
                1.0, -1.0, -1.0,
            };
            glBegin(GL_TRIANGLES_ADJACENCY_EXT);
            for(int i=0; i<6; i++) {
                glVertex3fv(v[i]);
            }
            glEnd();
            CHECK_ERRORS;
            if (!options[OPTION_USE_PROGRAM]) {
                // draw adjacent tris
                glColor3f(1.0, 0.0, 0.0);
                glBegin(GL_TRIANGLES);
                glVertex3fv(v[0]);
                glVertex3fv(v[1]);
                glVertex3fv(v[2]);

                glVertex3fv(v[2]);
                glVertex3fv(v[3]);
                glVertex3fv(v[4]);

                glVertex3fv(v[4]);
                glVertex3fv(v[5]);
                glVertex3fv(v[0]);
                glEnd();
            }
            CHECK_ERRORS;
            break;
        }

    case GL_LINES_ADJACENCY_EXT:
        {
            // 0 - - - 1----->2 - - - 3
            glBegin(GL_LINES_ADJACENCY_EXT);
            for(int i=0; i<nverts-1; i+=3) {
                glVertex2fv(v[i]);
                glVertex2fv(v[i+1]);
                glVertex2fv(v[i+2]);
                glVertex2fv(v[i+3]);
            }
            glEnd();
            CHECK_ERRORS;
            if (!options[OPTION_USE_PROGRAM]) {
                // make adjacent lines visible
                glColor3f(1.0, 0.0, 0.0);
                glBegin(GL_LINES);
                for(int i=0; i<nverts-1; i+=3) {
                    glVertex2fv(v[i]);
                    glVertex2fv(v[i+1]);
                    glVertex2fv(v[i+2]);
                    glVertex2fv(v[i+3]);
                }
                glEnd();
            }
            CHECK_ERRORS;
            break;
        }

    case GL_LINE_STRIP_ADJACENCY_EXT:
        {
            // 0 - - - 1--->2--->3--->4 - - - 5
            glBegin(GL_LINE_STRIP_ADJACENCY_EXT);
            for(int i=0; i<nverts; i++) {
                glVertex2fv(v[i]);
            }
            glEnd();
            CHECK_ERRORS;
            if (!options[OPTION_USE_PROGRAM]) {
                // make adjacent lines visible
                glColor3f(1.0, 0.0, 0.0);
                glBegin(GL_LINES);
                glVertex2fv(v[0]);
                glVertex2fv(v[1]);
                glVertex2fv(v[nverts-2]);
                glVertex2fv(v[nverts-1]);
                glEnd();
            }
            CHECK_ERRORS;
            break;
        }
	}
}

void draw_model(int mode)
{
	switch(mode) {
	case MODEL:
        if (model) {
            DrawTriangles( model);
        } 
        break;

    case MODEL_ADJ:
        if (model) {
            DrawWithAdjacency( model);
        } 
        break;

    case MODEL_POINTS:
        if (model) {
            DrawPoints( model);
        } 
        break;

    case MODEL_EDGES:
        if (model) {
            DrawEdges( model);
        } 
        break;
    }

    CHECK_ERRORS;
}

void draw_geometry(int npasses)
{
	for(int i=0; i<npasses; i++) {
	    CGparameter param = cgGetNamedParameter(current_prog->gprog, "passno");
		if (param) {
			cgSetParameter1i(param, i);
		}

		if (options[OPTION_DRAW_MODEL] && (current_prog->modelmode != MODEL_NONE)) {
			draw_model(current_prog->modelmode);
		} else {
			draw_prim(current_prog->prim);
		}
	}
}

//////////////////////////////////////////////////////////////////////
inline void updateButtonState( const nv::ButtonState &bs, nv::GlutManipulator &manip, int button) {
    int modMask = 0;

    if (bs.state & nv::ButtonFlags_Alt) modMask |= GLUT_ACTIVE_ALT;
    if (bs.state & nv::ButtonFlags_Shift) modMask |= GLUT_ACTIVE_SHIFT;
    if (bs.state & nv::ButtonFlags_Ctrl) modMask |= GLUT_ACTIVE_CTRL;

    if (bs.state & nv::ButtonFlags_End)
        manip.mouse( button, GLUT_UP, modMask, bs.cursor.x, height - bs.cursor.y);
    if (bs.state & nv::ButtonFlags_Begin)
        manip.mouse( button, GLUT_DOWN, modMask, bs.cursor.x, height - bs.cursor.y);
}

void doUI() {
    nv::Rect null;
    const char* programs[NPROGS];

    for (int ii=0; ii < NPROGS; ii++) {
        programs[ii] = prog[ii].name;
    }

    ui.begin();

	ui.beginGroup( nv::GroupFlags_GrowDownFromLeft);

    ui.doLabel( null, "Cg Geometry Program");
    if (ui.doComboBox( null, NPROGS, programs, &current_prog_idx)) {
        current_prog = &prog[current_prog_idx];
    }
    ui.doCheckButton( null, "Wireframe", &options[OPTION_DISPLAY_WIREFRAME]);
    ui.doCheckButton( null, "Use Model", &options[OPTION_DRAW_MODEL]);

    ui.doLabel( null, "Num Curve Segments");

    ui.beginGroup( nv::GroupFlags_GrowLeftFromTop);
    ui.doLabel( null, " ");
    float fsegments = segments;
    ui.doHorizontalSlider( null, 2.0f, 100.0f, &fsegments);
    segments = fsegments;
    ui.endGroup();

	ui.endGroup();

    // Pass non-ui mouse events to the manipulator
    if (!ui.isOnFocus()) {
        const nv::ButtonState &lbState = ui.getMouseState( 0);
        const nv::ButtonState &mbState = ui.getMouseState( 1);
        const nv::ButtonState &rbState =  ui.getMouseState( 2);

        manipulator.motion( ui.getCursorX(), height - ui.getCursorY());

        updateButtonState( lbState, manipulator, GLUT_LEFT_BUTTON);
        updateButtonState( mbState, manipulator, GLUT_MIDDLE_BUTTON);
        updateButtonState( rbState, manipulator, GLUT_RIGHT_BUTTON);
    }

    ui.end();
}

//////////////////////////////////////////////////////////////////////
void display()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    manipulator.applyTransform();
	
	// draw original object
    if (options[OPTION_OVERLAY_OBJECT] && (current_prog->modelmode != MODEL_NONE)) {
        glColor3f(0.0, 0.0, 0.0);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glPolygonOffset( 0.5f, 1.0f);
		glEnable(GL_POLYGON_OFFSET_FILL);
		cgGLBindProgram(prog[0].vprog);
		cgGLEnableProfile(cg_vprofile);
		draw_model(current_prog->modelmode);
		cgGLDisableProfile(cg_vprofile);
		glDisable(GL_POLYGON_OFFSET_FILL);
    }
    CHECK_ERRORS;

    glPolygonMode(GL_FRONT_AND_BACK, options[OPTION_DISPLAY_WIREFRAME] ? GL_LINE : GL_FILL);

	// bind programs
    if (options[OPTION_USE_PROGRAM]) {
		cgGLBindProgram(current_prog->vprog);
		cgGLEnableProfile(cg_vprofile);

		cgGLBindProgram(current_prog->gprog);
        cgGLEnableProfile(cg_gprofile);

	    CGparameter param = cgGetNamedParameter(current_prog->gprog, "segments");
		if (param) {
			cgGLSetParameter1f(param, segments);
			// set geometry program output size (for performance)
			glProgramVertexLimitNV(GL_GEOMETRY_PROGRAM_NV, segments);
		}

		if (current_prog->fprog) {
			cgGLBindProgram(current_prog->fprog);
			cgGLEnableProfile(cg_fprofile);
			CGparameter param = cgGetNamedParameter(current_prog->fprog, "viewport");
			if (param) {
				cgGLSetParameter2f(param, width, height);
			}
		}
	}

    CHECK_ERRORS;

    if (enableBlending) {
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        glEnable(GL_BLEND);
    }
    if (enableDepthTest) {
        glEnable(GL_DEPTH_TEST);
    }

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_RECTANGLE_NV, rand_tex);

    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_3D, noise_tex);

//    glColor3f(0.1, 0.1, 0.1);
    glColor3f(1.0, 1.0, 1.0);
	if (current_prog->multipass) {
		draw_geometry(passes);
	} else {
		draw_geometry(1);
	}

    glDisable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);

    cgGLDisableProfile(cg_vprofile);
    cgGLDisableProfile(cg_gprofile);
    cgGLDisableProfile(cg_fprofile);

    CHECK_ERRORS;

    glutReportErrors();

    doUI();

    glutSwapBuffers();
}


//////////////////////////////////////////////////////////////////////
void idle()
{
    if ( options[OPTION_ANIMATE]) {
        manipulator.idle();
    }

    glutPostRedisplay();
}


//////////////////////////////////////////////////////////////////////
void key(unsigned char k, int x, int y)
{
    ui.keyboard( k, x, y);
	k = tolower(k);

    if (optionKeyMap.find(k) != optionKeyMap.end())
        options[optionKeyMap[k]] = !options[optionKeyMap[k]];

    switch(k) {
        case 27:
        case 'q':
            exit(0);
            break;

        case '=':
        case '+':
            if (segments < 100) segments++;
            break;

        case '-':
            if (segments > 2) segments--;
            break;

		case ']':
			passes++;
			break;
		case '[':
			if (passes > 1) passes--;
			break;
        case 'b':
            enableBlending ^= 1;
            break;
        case 'd':
            enableDepthTest ^= 1;
            break;

        case '1':
            current_prog = &prog[0];
            current_prog_idx = 0;
            break;
        case '2':
            current_prog = &prog[NPROGS-1];
            current_prog_idx = NPROGS-1;
            break;
    }

    
	glutPostRedisplay();
}


//////////////////////////////////////////////////////////////////////
void resize(int w, int h)
{
    if (h == 0) h = 1;

    ui.reshape( w, h);

    glViewport(0, 0, w, h);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 0.1, 100.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    manipulator.reshape(w, h);

    width = w;
    height = h;
}


//////////////////////////////////////////////////////////////////////
void mouse(int button, int state, int x, int y)
{
    ui.mouse( button, state, glutGetModifiers(), x, y);
}


//////////////////////////////////////////////////////////////////////
void motion(int x, int y)
{
    ui.mouseMotion( x, y);
}

//////////////////////////////////////////////////////////////////////
void main_menu(int i)
{
  key((unsigned char) i, 0, 0);
}

//////////////////////////////////////////////////////////////////////
void prog_menu(int i)
{
    current_prog = &prog[i];
    current_prog_idx = i;
}


//////////////////////////////////////////////////////////////////////
void init_menus()
{
    int prog_menu_id = glutCreateMenu(prog_menu);
    for(int i=0; i<NPROGS; i++) {
        glutAddMenuEntry( prog[i].name, i);
    }

    glutCreateMenu(main_menu);
    glutAddSubMenu("Program", prog_menu_id);
    glutAddMenuEntry("Toggle geometry program [p]", 'p');
    glutAddMenuEntry("Toggle model / primitive [m]", 'm');
    glutAddMenuEntry("Toggle wireframe [w]", 'w');
    glutAddMenuEntry("Toggle original model [o]", 'o');
    glutAddMenuEntry("Increase tessellation [=]", '=');
    glutAddMenuEntry("Decrease tessellation [-]", '-');
    glutAddMenuEntry("Quit (esc)", '\033');
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}


//////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitWindowSize( width, height);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);

	glutCreateWindow("Cg geometry programs");

	init_opengl();

    // configure a simple controller
    manipulator.setDollyActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_CTRL);
    manipulator.setPanActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_SHIFT);
    manipulator.setDollyPosition( -2.0f);

    // setup event callback functions
	glutDisplayFunc(display);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutPassiveMotionFunc(motion);
    glutIdleFunc(idle);
    glutKeyboardFunc(key);
    glutReshapeFunc(resize);
    init_menus();

    // accept alternate models via the command line
    if (argc > 1) {
        model_filename = argv[1];
    }

    // create the model and make it ready for display
    model = new nv::Model;
    std::string resolved_path;
    if ( sdkPath.getFilePath( model_filename, resolved_path) ) {
        if (model->loadModelFromFile(resolved_path.c_str())) {

            // remove any primitives with duplicate indices, as they might confuse adjacency
            model->removeDegeneratePrims();

            // compute normal and tengent space
            model->computeNormals();
            model->computeTangents();

            // make the model efficient for rendering with vertex arrays
            model->compileModel( nv::Model::eptAll);

            // get the bounding box to help place the model on-screen
            model->computeBoundingBox( modelMin, modelMax);
        } else {
            fprintf(stderr, "Error loading model '%s'\n", model_filename);
            delete model;
            model = 0;
        }
    }
    else {
        fprintf( stderr, "Couldn't find model '%s'\n", model_filename);
        model = 0;
    }

    //configure the options
    optionKeyMap['w'] = OPTION_DISPLAY_WIREFRAME;
    options[OPTION_DISPLAY_WIREFRAME] = false;

    optionKeyMap['p'] = OPTION_USE_PROGRAM;
    options[OPTION_USE_PROGRAM] = true;
    
    optionKeyMap[' '] = OPTION_ANIMATE;
    options[OPTION_ANIMATE] = true;

    optionKeyMap['o'] = OPTION_OVERLAY_OBJECT;
    options[OPTION_OVERLAY_OBJECT] = false;

    optionKeyMap['m'] = OPTION_DRAW_MODEL;
    options[OPTION_DRAW_MODEL] = true;

    //print the help info
    printf( "cg_geometry_program - sample showing the usage of geometry programs with Cg\n");
    printf( "  Commands:\n");
    printf( "    q / [ESC] - Quit the application\n");
    printf( "    [SPACE]   - Toggle continuous animation\n");
    printf( "    p         - Toggle using a Cg program\n");
    printf( "    m         - Toggle model / primitive\n");
    printf( "    w         - Toggle displaying wireframe\n");
    printf( "    o         - Toggle overlaying the result on the object\n");
    printf( "    =         - Increase tessellation\n");
    printf( "    -         - Decrease tessellation\n\n");

	glutMainLoop();

	return 0;
}